@@ -1,5 +1,5 @@
 from .agent import DerivationAgent, FlaskConfig
-from .conf import AgentConfig, config_derivation_agent
+from .conf import AgentConfig, config_derivation_agent, Config, config_generic
 from .data_model import DerivationInputs, DerivationOutputs
 
-__version__ = "1.0.0"
+__version__ = "1.1.0"
