@@ -1,7 +1,10 @@
+from abc import ABC, abstractmethod
+from typing import Type, TypeVar
 from flask_apscheduler import APScheduler
 from flask import Flask
 from flask import request
 from urllib.parse import unquote
+from urllib.parse import urlparse
 import json
 import time
 
@@ -10,6 +13,9 @@ import agentlogging
 from pyderivationagent.kg_operations import *
 from pyderivationagent.data_model import DerivationInputs, DerivationOutputs
 
+# see https://mypy.readthedocs.io/en/latest/generics.html#type-variable-upper-bound
+PY_SPARQL_CLIENT = TypeVar('PY_SPARQL_CLIENT', bound=PySparqlClient)
+
 
 class FlaskConfig(object):
     """
@@ -18,7 +24,7 @@ class FlaskConfig(object):
     SCHEDULER_API_ENABLED = True
 
 
-class DerivationAgent(object):
+class DerivationAgent(ABC):
     def __init__(
         self,
         agent_iri: str,
@@ -33,7 +39,8 @@ class DerivationAgent(object):
         fs_password: str = None,
         app: Flask = Flask(__name__),
         flask_config: FlaskConfig = FlaskConfig(),
-        agent_endpoint: str = "/",
+        agent_endpoint: str = "http://localhost:5000/sync_derivation",
+        register_agent: bool = True,
         logger_name: str = "dev"
     ):
         """
@@ -42,7 +49,7 @@ class DerivationAgent(object):
             Arguments:
                 app - flask app object, an example: app = Flask(__name__)
                 agent_iri - OntoAgent:Service IRI of the derivation agent, an example: "http://www.example.com/triplestore/agents/Service__XXXAgent#Service"
-                agent_endpoint - data property OntoAgent:hasHttpUrl of OntoAgent:Operation of the derivation agent, an example: "http://localhost:7000/endpoint"
+                agent_endpoint - data property OntoAgent:hasHttpUrl of OntoAgent:Operation of the derivation agent, an example: "http://localhost:5000/endpoint"
                 time_interval - time interval between two runs of derivation monitoring job (in SECONDS)
                 derivation_instance_base_url - namespace to be used when creating derivation instance, an example: "http://www.example.com/triplestore/repository/"
                 kg_url - SPARQL query endpoint, an example: "http://localhost:8080/blazegraph/namespace/triplestore/sparql"
@@ -53,6 +60,7 @@ class DerivationAgent(object):
                 fs_user - username used to access the file server endpoint specified by fs_url
                 fs_password - password that set for the fs_user used to access the file server endpoint specified by fs_url
                 flask_config - configuration object for flask app, should be an instance of the class FlaskConfig provided as part of this package
+                register_agent - boolean value, whether to register the agent to the knowledge graph
                 logger_name - logger names for getting correct loggers from agentlogging package, valid logger names: "dev" and "prod", for more information, visit https://github.com/cambridge-cares/TheWorldAvatar/blob/develop/Agents/utils/python-utils/agentlogging/logging.py
         """
 
@@ -67,7 +75,7 @@ class DerivationAgent(object):
         self.app.config.from_object(flask_config)
 
         # initialise flask scheduler and assign time interval for monitorDerivations job
-        self.scheduler = APScheduler()
+        self.scheduler = APScheduler(app=self.app)
         self.time_interval = time_interval
 
         # assign IRI and HTTP URL of the agent
@@ -95,18 +103,80 @@ class DerivationAgent(object):
         self.derivationClient = self.jpsBaseLib_view.DerivationClient(
             self.storeClient, derivation_instance_base_url)
 
+        # initialise the SPARQL client as None, this will be replaced when get_sparql_client() is first called
+        self.sparql_client = None
+
+        # initialise the logger
         self.logger = agentlogging.get_logger(logger_name)
+
+        # register the agent to the KG if required
+        self.register_agent = register_agent
+        try:
+            self.register_agent_in_kg()
+        except Exception as e:
+            self.logger.error(
+                "Failed to register the agent <{}> to the KG <{}>. Error: {}".format(self.agentIRI, self.kgUrl, e),
+                stack_info=True, exc_info=True)
+            raise e
+
         self.logger.info(
             "DerivationAgent <%s> is initialised to monitor derivations in triple store <%s> with a time interval of %d seconds." % (
                 self.agentIRI, self.kgUrl, self.time_interval)
         )
 
+    def periodical_job(func):
+        """This method is used to start a periodic job. This should be used as a decorator (@Derivation.periodical_job) for the method that needs to be executed periodically."""
+        def inner(self, *args, **kwargs):
+            func(self, *args, **kwargs)
+            if not self.scheduler.running:
+                self.scheduler.start()
+                self.logger.info("Scheduler is started.")
+        inner.__is_periodical_job__ = True
+        return inner
+
+    def get_sparql_client(self, sparql_client_cls: Type[PY_SPARQL_CLIENT]) -> PY_SPARQL_CLIENT:
+        """This method returns a SPARQL client object that instantiated from sparql_client_cls, which should extend PySparqlClient class."""
+        if self.sparql_client is None or not isinstance(self.sparql_client, sparql_client_cls):
+            self.sparql_client = sparql_client_cls(
+                query_endpoint=self.kgUrl, update_endpoint=self.kgUpdateUrl,
+                kg_user=self.kgUser, kg_password=self.kgPassword,
+                fs_url=self.fs_url, fs_user=self.fs_user, fs_pwd=self.fs_password
+            )
+        return self.sparql_client
+
+    def register_agent_in_kg(self):
+        """This method registers the agent to the knowledge graph by uploading its OntoAgent triples generated on-the-fly."""
+        if self.register_agent:
+            sparql_client = self.get_sparql_client(PySparqlClient)
+            input_concepts = self.agent_input_concepts()
+            output_concepts = self.agent_output_concepts()
+            if not isinstance(input_concepts, list) or not isinstance(output_concepts, list):
+                raise Exception("Failed to register the agent <{}> to the KG <{}>. Error: Input and output concepts must be lists. Received: {} (type: {}) and {} (type: {})".format(
+                    self.agentIRI, self.kgUrl, input_concepts, type(input_concepts), output_concepts, type(output_concepts)))
+            if len(input_concepts) == 0 or len(output_concepts) == 0:
+                raise Exception("Failed to register the agent <{}> to the KG <{}>. Error: No input or output concepts specified.".format(self.agentIRI, self.kgUrl))
+            sparql_client.generate_ontoagent_instance(self.agentIRI, self.agentEndpoint, input_concepts, output_concepts)
+            self.logger.info("Agent <%s> is registered to the KG <%s> with input signature %s and output signature %s." % (
+                self.agentIRI, self.kgUrl, input_concepts, output_concepts))
+        else:
+            self.logger.info("Flag register_agent is False. Agent <%s> is NOT registered to the KG <%s>." % (self.agentIRI, self.kgUrl))
+
+    @abstractmethod
+    def agent_input_concepts(self) -> list:
+        """This method returns a list of input concepts of the agent. This should be overridden by the derived class."""
+        pass
+
+    @abstractmethod
+    def agent_output_concepts(self) -> list:
+        """This method returns a list of output concepts of the agent. This should be overridden by the derived class."""
+        pass
+
     def add_url_pattern(self, url_pattern=None, url_pattern_name=None, function=None, methods=['GET'], *args, **kwargs):
         """
             This method is a wrapper of add_url_rule method of Flask object that adds customised URL Pattern to derivation agent.
-            #flask.Flask.add_url_rule.
-            For more information, visit https://flask.palletsprojects.com/en/2.0.x/api/
-            WARNING: Use of this is STRONGLY discouraged. The design intention of an derivation agent is to communicate via the KNOWLEDGE GRAPH, and NOT via HTTP requests.
+            For more information, visit https://flask.palletsprojects.com/en/2.0.x/api/#flask.Flask.add_url_rule
+            WARNING: Use of this by developer is STRONGLY discouraged.
+            The design intention of an derivation agent is to communicate via the KNOWLEDGE GRAPH, and NOT via HTTP requests.
 
             Arguments:
                 url_pattern - the endpoint url to associate with the rule and view function
@@ -217,6 +287,7 @@ class DerivationAgent(object):
         else:
             self.logger.info("Currently, no asynchronous derivation <isDerivedUsing> <%s>." % (self.agentIRI))
 
+    @abstractmethod
     def process_request_parameters(self, derivation_inputs: DerivationInputs, derivation_outputs: DerivationOutputs):
         """
         This method perform the agent logic of converting derivation inputs to derivation outputs.
@@ -236,40 +307,68 @@ class DerivationAgent(object):
         """
         pass
 
-    def start_monitoring_derivations(self):
+    @periodical_job
+    def _start_monitoring_derivations(self):
         """
-            This method starts the periodical job to monitor derivation, also runs the flask app as an HTTP servlet.
+            This method starts the periodical job to monitor asynchronous derivation, also adds the HTTP endpoint to handle synchronous derivation.
         """
-        self.scheduler.init_app(self.app)
         self.scheduler.add_job(id='monitor_derivations', func=self.monitor_async_derivations,
                                trigger='interval', seconds=self.time_interval)
-        self.scheduler.start()
-        self.logger.info("Monitor asynchronous derivations job is started with a time interval of %d seconds." % (
+        self.logger.info("Monitor asynchronous derivations job is scheduled with a time interval of %d seconds." % (
             self.time_interval))
 
-        self.add_url_pattern(self.agentEndpoint, self.agentEndpoint[1:], self.handle_sync_derivations, methods=['GET'])
+        url_pattern = urlparse(self.agentEndpoint).path
+        url_pattern_name = url_pattern.strip('/').replace('/', '_') + '_handle_sync_derivations'
+        self.add_url_pattern(url_pattern, url_pattern_name, self.handle_sync_derivations, methods=['GET'])
         self.logger.info("Synchronous derivations can be handled at endpoint: " + self.agentEndpoint)
 
+    def start_all_periodical_job(self):
+        """This method starts all scheduled periodical jobs."""
+        all_periodical_jobs = [getattr(self, name) for name in dir(self) if callable(getattr(self, name)) and not name.startswith('__') and hasattr(getattr(self, name), '__is_periodical_job__')]
+        for func in all_periodical_jobs:
+            func()
+
     def handle_sync_derivations(self):
-        requestParams = json.loads(unquote(request.url[len(request.base_url):])[
-            len("?query="):])
+        self.logger.info("Received synchronous derivation request: %s." % (request.url))
+        requestParams = json.loads(unquote(urlparse(request.url).query)[len("query="):])
         res = {}
         if self.validate_inputs(requestParams):
+            # serialises DerivationInputs objects from JSONObject
             inputs = self.jpsBaseLib_view.DerivationInputs(requestParams[self.jpsBaseLib_view.DerivationClient.AGENT_INPUT_KEY])
             self.logger.info("Received derivation request parameters: " + str(requestParams))
+
+            # retrieve necessary information
             derivationIRI = requestParams[self.jpsBaseLib_view.DerivationClient.DERIVATION_KEY]
             derivationType = requestParams[self.jpsBaseLib_view.DerivationClient.DERIVATION_TYPE_KEY]
+            syncNewInfoFlag = requestParams[self.jpsBaseLib_view.DerivationClient.SYNC_NEW_INFO_FLAG]
+
+            # initialise DerivationOutputs, also set up information
             outputs = self.jpsBaseLib_view.DerivationOutputs()
             outputs.setThisDerivation(derivationIRI)
             outputs.setRetrievedInputsAt(int(time.time()))
-            outputs.setOldEntitiesMap(requestParams[self.jpsBaseLib_view.DerivationClient.BELONGSTO_KEY])
-            outputs.setOldEntitiesDownstreamDerivationMap(requestParams[self.jpsBaseLib_view.DerivationClient.DOWNSTREAMDERIVATION_KEY])
+            if not syncNewInfoFlag:
+                outputs.setOldEntitiesMap(requestParams[self.jpsBaseLib_view.DerivationClient.BELONGSTO_KEY])
+                outputs.setOldEntitiesDownstreamDerivationMap(requestParams[self.jpsBaseLib_view.DerivationClient.DOWNSTREAMDERIVATION_KEY])
 
             # apply agent logic to convert inputs to outputs
             derivation_inputs = DerivationInputs(inputs)
             derivation_outputs = DerivationOutputs(outputs)
             self.process_request_parameters(derivation_inputs, derivation_outputs)
 
+            # return response if this sync derivation is generated for new info
+            if syncNewInfoFlag:
+                agentServiceIRI = requestParams[self.jpsBaseLib_view.DerivationClient.AGENT_IRI_KEY]
+                self.derivationClient.writeSyncDerivationNewInfo(
+                    outputs.getOutputTriples(), outputs.getNewDerivedIRI(),
+                    agentServiceIRI, inputs.getAllIris(),
+                    derivationIRI, derivationType, outputs.getRetrievedInputsAt()
+                )
+                res[self.jpsBaseLib_view.DerivationOutputs.RETRIEVED_INPUTS_TIMESTAMP_KEY] = outputs.getRetrievedInputsAt()
+                res[self.jpsBaseLib_view.DerivationClient.AGENT_OUTPUT_KEY] = json.loads(str(outputs.getNewEntitiesJsonMap()))
+                self.logger.info("Synchronous derivation for new information generated successfully, returned response: " + str(res))
+                return json.dumps(res)
+
+            # only enters below if the computation was not for new information (new instances)
             derivation = self.jpsBaseLib_view.Derivation(derivationIRI, derivationType)
             if not derivation.isDerivationAsyn() and not derivation.isDerivationWithTimeSeries():
                 # construct and fire SPARQL update given DerivationOutputs objects, if normal
@@ -308,11 +407,41 @@ class DerivationAgent(object):
 
         return json.dumps(res)
 
+    @abstractmethod
     def validate_inputs(self, http_request) -> bool:
+        # developer can overwrite this function for customised validation
+        self.logger.info("Validating inputs: " + str(http_request))
+        if not bool(http_request):
+            self.logger.warn("RequestParams are empty, throwing BadRequestException...")
+            raise Exception("RequestParams are empty")
+
+        if self.jpsBaseLib_view.DerivationClient.AGENT_INPUT_KEY not in http_request:
+            self.logger.info("Agent <%s> received an empty request..." % self.agentIRI)
+            return False
+        else:
+            if self.jpsBaseLib_view.DerivationClient.DERIVATION_KEY not in http_request:
+                msg = "Agent <%s> received a request that doesn't have derivationIRI..." % self.agentIRI
+                self.logger.error(msg)
+                raise Exception(msg)
+            if http_request[self.jpsBaseLib_view.DerivationClient.SYNC_NEW_INFO_FLAG]:
+                if self.jpsBaseLib_view.DerivationClient.AGENT_IRI_KEY not in http_request:
+                    msg = "Agent <%s> received a request for sync new information that doesn't have information about agent IRI..." % self.agentIRI
+                    self.logger.error(msg)
+                    raise Exception(msg)
+            else:
+                if self.jpsBaseLib_view.DerivationClient.BELONGSTO_KEY not in http_request:
+                    msg = "Agent <%s> received a request that doesn't have information about old outputs..." % self.agentIRI
+                    self.logger.error(msg)
+                    raise Exception(msg)
+                if self.jpsBaseLib_view.DerivationClient.DOWNSTREAMDERIVATION_KEY not in http_request:
+                    msg = "Agent <%s> received a request that doesn't have information about downstream derivation..." % self.agentIRI
+                    self.logger.error(msg)
+                    raise Exception(msg)
+
         return True
 
     def run_flask_app(self, **kwargs):
         """
-            This method starts the periodical job to monitor derivation, also runs the flask app as an HTTP servlet.
+            This method runs the flask app as an HTTP servlet.
         """
         self.app.run(**kwargs)
